#pragma once

#include <list>

#include <amdis/Output.hpp>
#include <amdis/linearalgebra/DOFVectorInterface.hpp>

namespace AMDiS
{
  class GridTransferInterface
  {
  public:
    virtual ~GridTransferInterface() = default;

    virtual void attach(DOFVectorInterface*) = 0;
    virtual void detach(DOFVectorInterface*) = 0;
    virtual bool preAdapt() = 0;
    virtual bool adapt() = 0;
    virtual void postAdapt() = 0;

  };


  template <class Grid>
  class GridTransfer
      : public GridTransferInterface
  {
    using Self = GridTransfer;

  public:
    void bind(Grid& grid)
    {
      grid_ = &grid;
    }

    void unbind()
    {
      grid_ = nullptr;
    }

    /// Attach a data container to the grid transfer, that gets interpolated during grid change
    void attach(DOFVectorInterface* vec) override
    {
      data_.push_back(vec);
    }

    void detach(DOFVectorInterface* vec) override
    {
      auto it = std::find(data_.begin(), data_.end(), vec);
      if (it != data_.end())
        data_.erase(it);
      else
        warning("DOFVector to detach not found");
    }

    /// Prepare the grid and the data for the adaption
    bool preAdapt() override
    {
      assert(grid_ != nullptr);
      mightCoarsen_ = grid_->preAdapt(); // any element might be coarsened in adapt()
      for (auto* vec : this->data_)
        vec->preAdapt(mightCoarsen_);
      return mightCoarsen_;
    }

    /// do the grid adaption
    bool adapt() override
    {
      assert(grid_ != nullptr);
      refined_ = grid_->adapt(); // returns true if a least one entity was refined
      return refined_;
    }

    // Perform data adaption to the new grid
    void postAdapt() override
    {
      assert(grid_ != nullptr);
      if (mightCoarsen_ || refined_) {
        for (auto* vec : this->data_)
          vec->postAdapt(refined_);
      }
      grid_->postAdapt();
      changeIndex_++;
    }

    /// Returns the grid change index, see \ref changeIndex.
    unsigned long changeIndex() const
    {
      return changeIndex_;
    }

  private:
    Grid* grid_ = nullptr;
    std::list<DOFVectorInterface*> data_;
    bool mightCoarsen_ = false;
    bool refined_ = false;

    /// This index is incremented every time the grid is changed, e.g. by
    /// refinement or coarsening.
    unsigned long changeIndex_ = 0;
  };

} // end namespace AMDiS
