#pragma once

#include <tuple>
#include <type_traits>

#include <dune/common/ftraits.hh>

#include <amdis/common/Concepts.hpp>
#include <amdis/common/Loops.hpp>
#include <amdis/common/Math.hpp>
#include <amdis/common/Mpl.hpp>
#include <amdis/common/MultiTypeVector.hpp>
#include <amdis/common/StaticSize.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/typetree/MultiIndex.hpp>

namespace AMDiS
{
  // forward declaration
  template <class... Rows>
  class MultiTypeMatrix;
}

namespace Dune
{
  template <class... Rows>
  struct FieldTraits<AMDiS::MultiTypeMatrix<Rows...>>
  {
    using field_type = std::common_type_t<typename Dune::FieldTraits<Rows>::field_type...>;
    using real_type = std::common_type_t<typename Dune::FieldTraits<Rows>::real_type...>;
  };
}


namespace AMDiS
{
  // Rows should be of type MultiTypeVector
  template <class... Rows>
  class MultiTypeMatrix
      : public std::tuple<Rows...>
  {
    using Self = MultiTypeMatrix;
    using Super = std::tuple<Rows...>;

    static_assert(IsEqual<int, Rows::dimension...>::value,
      "All columns must have the same length.");

  public:
    using field_type = typename Dune::FieldTraits<Self>::field_type;
    using real_type = typename Dune::FieldTraits<Self>::real_type;
    using size_type = std::size_t;

    static constexpr int rows = std::tuple_size<Super>::value;
    static constexpr int cols = Math::max(Rows::dimension...);

    template <class... Rows_,
      REQUIRES( Concepts::Similar<Types<Rows...>, Types<Rows_...>> )>
    MultiTypeMatrix(Rows_&&... rows)
      : Super(std::forward<Rows_>(rows)...)
    {}

    /// Default construction of tuple of FieldVectors
    MultiTypeMatrix() = default;

    /// Construct tuple by initializing all tuple elements with a constant value
    explicit MultiTypeMatrix(real_type value)
    {
      *this = value;
    }

    /// Assignment of real number to all tuple elements
    MultiTypeMatrix& operator=(real_type value)
    {
      forEach(*this, [value](auto& fv) { fv = value; });
      return *this;
    }

    // Compound assignment operator +=
    MultiTypeMatrix& operator+=(MultiTypeMatrix const& that)
    {
      forEach(range_<0,rows>, [&that,this](auto const _i) { (*this)[_i] += that[_i]; });
      return *this;
    }

    // Compound assignment operator -=
    MultiTypeMatrix& operator-=(MultiTypeMatrix const& that)
    {
      forEach(range_<0,rows>, [&that,this](auto const _i) { (*this)[_i] -= that[_i]; });
      return *this;
    }

    // Scaling of all tuple elements by a constant value
    MultiTypeMatrix& operator*=(real_type value)
    {
      forEach(*this, [value](auto& fv) { fv *= value; });
      return *this;
    }

    // Scaling of all tuple elements by the inverse of a constant value
    MultiTypeMatrix& operator/=(real_type value)
    {
      forEach(*this, [value](auto& fv) { fv /= value; });
      return *this;
    }

    /// Const access to the tuple elements
    template <std::size_t I, std::size_t J>
    decltype(auto) operator()(const index_t<I>&, const index_t<J>&) const
    {
      return std::get<J>(std::get<I>(*this));
    }

    /// Mutable access to the tuple elements
    template <std::size_t I, std::size_t J>
    decltype(auto) operator()(const index_t<I>&, const index_t<J>&)
    {
      return std::get<J>(std::get<I>(*this));
    }


    /// Const access to the rows
    template <std::size_t I>
    decltype(auto) operator[](const index_t<I>&) const
    {
      return std::get<I>(*this);
    }

    /// Mutable access to the rows
    template <std::size_t I>
    decltype(auto) operator[](const index_t<I>&)
    {
      return std::get<I>(*this);
    }

    /// Return number of elements of the tuple
    static constexpr std::size_t num_rows()
    {
      return rows;
    }

    /// Return number of elements of the tuple
    static constexpr std::size_t num_cols()
    {
      return cols;
    }

    static constexpr std::size_t size()
    {
      return rows;
    }
  };

} // end namespace AMDiS
