#pragma once

#include <array>
#include <tuple>
#include <type_traits>

namespace Dune
{
  // forward declarations
  template <class T, int N, int M>
  class FieldMatrix;

  template <class T, int N>
  class FieldVector;

  template <class... Ts>
  class TupleVector;

  template <class Row0, class... Rows>
  class MultiTypeBlockMatrix;

  template <class... Ts>
  class MultiTypeBlockVector;
}

namespace AMDiS
{
  // forward declarations
  template <class... Rows>
  class MultiTypeMatrix;

  template <class... Ts>
  class MultiTypeVector;


  namespace Impl
  {
    template <class Tuple, class = void>
    struct SizeImpl : std::integral_constant<std::size_t, 0> {};

    template <class... Args>
    struct SizeImpl<std::tuple<Args...>>
        : std::integral_constant<std::size_t, sizeof...(Args) > {};

    template <class Arg0, class Arg1>
    struct SizeImpl<std::pair<Arg0, Arg1>>
        : std::integral_constant<std::size_t, 2> {};

    template <class T, std::size_t N>
    struct SizeImpl<std::array<T,N>>
        : std::integral_constant<std::size_t, N> {};

    template <class T, int N>
    struct SizeImpl<Dune::FieldVector<T,N>>
        : std::integral_constant<std::size_t, N> {};

    template <class T, int N, int M>
    struct SizeImpl<Dune::FieldMatrix<T,N,M>>
        : std::integral_constant<std::size_t, N*M> {};

    template <class... Ts>
    struct SizeImpl<MultiTypeVector<Ts...>>
        : std::integral_constant<std::size_t, sizeof...(Ts) > {};

    template <class... Ts>
    struct SizeImpl<Dune::TupleVector<Ts...>>
        : std::integral_constant<std::size_t, sizeof...(Ts)> {};

    template <class... Ts>
    struct SizeImpl<Dune::MultiTypeBlockVector<Ts...>>
        : std::integral_constant<std::size_t, sizeof...(Ts)> {};

    // Specialization for arithmetic types
    template <class T>
    struct SizeImpl<T, std::enable_if_t<std::is_arithmetic<T>::value> >
        : std::integral_constant<std::size_t, 1> {};

  } // end namespace Impl

  /// Get the number of elements in a tuple / pair / array / ...
  template <class T>
  constexpr std::size_t Size_v = Impl::SizeImpl<std::decay_t<T>>::value;

  template <class T>
  using Size_t = Impl::SizeImpl<std::decay_t<T>>;


  namespace Impl
  {
    template <class Tuple, class = void>
    struct RowsImpl : std::integral_constant<std::size_t, 0> {};

    template <class T, int N, int M>
    struct RowsImpl<Dune::FieldMatrix<T,N,M>>
        : std::integral_constant<std::size_t, N> {};

    template <class... Rows>
    struct RowsImpl<MultiTypeMatrix<Rows...>>
        : std::integral_constant<std::size_t, sizeof...(Rows) > {};

    template <class... Rows>
    struct RowsImpl<Dune::MultiTypeBlockMatrix<Rows...>>
        : std::integral_constant<std::size_t, sizeof...(Rows)> {};

    // Specialization for arithmetic types
    template <class T>
    struct RowsImpl<T, std::enable_if_t<std::is_arithmetic<T>::value> >
        : std::integral_constant<std::size_t, 1> {};

  } // end namespace Impl

  /// Get the number of rows in a fixed-size matrix
  template <class T>
  constexpr std::size_t Rows_v = Impl::RowsImpl<std::decay_t<T>>::value;

  template <class T>
  using Rows_t = Impl::RowsImpl<std::decay_t<T>>;


  namespace Impl
  {
    template <class Tuple, class = void>
    struct ColsImpl : std::integral_constant<std::size_t, 0> {};

    template <class T, int N, int M>
    struct ColsImpl<Dune::FieldMatrix<T,N,M>>
        : std::integral_constant<std::size_t, M> {};

    template <class Row0, class... Rows>
    struct ColsImpl<MultiTypeMatrix<Row0, Rows...>>
        : SizeImpl<Row0> {};

    template <class Row0, class... Rows>
    struct ColsImpl<Dune::MultiTypeBlockMatrix<Row0, Rows...>>
        : SizeImpl<Row0> {};

    // Specialization for arithmetic types
    template <class T>
    struct ColsImpl<T, std::enable_if_t<std::is_arithmetic<T>::value> >
        : std::integral_constant<std::size_t, 1> {};

  } // end namespace Impl

  /// Get the number of columns in a fixed-size matrix
  template <class T>
  constexpr std::size_t Cols_v = Impl::ColsImpl<std::decay_t<T>>::value;

  template <class T>
  using Cols_t = Impl::ColsImpl<std::decay_t<T>>;

} // end namespace AMDiS
