#pragma once

#include <algorithm>
#include <string>

#include <amdis/linearalgebra/RunnerInterface.hpp>
#include <amdis/linearalgebra/SolverInfo.hpp>
#include <amdis/linearalgebra/eigen/SolverConfig.hpp>

namespace AMDiS
{
  /**
   * \ingroup Solver
   * \class AMDiS::DirectRunner
   * \brief \implements RunnerInterface for the (external) direct solvers
   */
  template <class LU, class Matrix, class VectorX, class VectorB>
  class DirectRunner
      : public RunnerInterface<Matrix, VectorX, VectorB>
  {
  protected:
    using Super = RunnerInterface<Matrix, VectorX, VectorB>;

  public:
    /// Constructor.
    DirectRunner(std::string const& prefix)
      : solver_{}
    {
      SolverConfig<LU>::init(prefix, solver_);
      Parameters::get(prefix + "->reuse pattern", reusePattern_);
    }

    /// Implementes \ref RunnerInterface::init()
    void init(Matrix const& A) override
    {
      if (!reusePattern_ || !initialized_) {
        solver_.analyzePattern(A);
        initialized_ = true;
      }
      solver_.factorize(A);

      test_exit(solver_.info() == Eigen::Success,
        "Error in solver.compute(matrix)");
    }


    /// Implementes \ref RunnerInterface::exit()
    void exit() override {}

    /// Implementes \ref RunnerInterface::solve()
    int solve(Matrix const& A, VectorX& x, VectorB const& b,
              SolverInfo& solverInfo) override
    {
      x = solver_.solve(b);

      auto r = VectorB(b);
      if (x.norm() != 0)
        r -= A * x;

      solverInfo.setAbsResidual(r.norm());
      solverInfo.setError(solver_.info());

      return solver_.info() == Eigen::Success ? 0 : 1;
    }

  private:
    LU solver_;
    bool reusePattern_ = false;
    bool initialized_ = false;
  };
}
