#pragma once

#include <type_traits>

#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/LocalOperator.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  /// convection-diffusion operator, see \ref convectionDiffusion
  /// <A*grad(u),grad(v)> - <b*u, grad(v)> + <c*u, v> = <f, v> (conserving) or
  /// <A*grad(u),grad(v)> + <b*grad(u), v> + <c*u, v> = <f, v> (non conserving)
  template <class LocalContext, class GridFctA, class GridFctB, class GridFctC, class GridFctF, bool conserving = true>
  class ConvectionDiffusionOperator
      : public LocalOperator<ConvectionDiffusionOperator<LocalContext, GridFctA, GridFctB, GridFctC, GridFctF, conserving>,
                             LocalContext>
  {
    static const int dow = ContextGeometry<LocalContext>::dow;

    using A_range = typename GridFctA::Range;
    static_assert( Size_v<A_range> == 1 || (Rows_v<A_range> == dow && Cols_v<A_range> == dow),
      "Expression A(x) must be of scalar or matrix type." );
    using b_range = typename GridFctB::Range;
    static_assert( Size_v<b_range> == 1 || Size_v<b_range> == dow,
      "Expression b(x) must be of scalar or vector type." );
    using c_range = typename GridFctC::Range;
    static_assert( Size_v<c_range> == 1,
      "Expression c(x) must be of scalar type." );
    using f_range = typename GridFctF::Range;
    static_assert( Size_v<f_range> == 1,
      "Expression f(x) must be of scalar type." );

  public:
    ConvectionDiffusionOperator(GridFctA const& gridFctA, GridFctB const& gridFctB,
                                GridFctC const& gridFctC, GridFctF const& gridFctF)
      : gridFctA_(gridFctA)
      , gridFctB_(gridFctB)
      , gridFctC_(gridFctC)
      , gridFctF_(gridFctF)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode, ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isLeaf,
        "Operator can be applied to Leaf-Nodes only." );

      static_assert(std::is_same<FiniteElementType_t<RowNode>, FiniteElementType_t<ColNode>>{},
        "Galerkin operator requires equal finite elements for test and trial space." );

      using RangeFieldType = typename NodeQuadCache<RowNode>::Traits::RangeFieldType;

      auto localFctA = localFunction(gridFctA_); localFctA.bind(context.element());
      auto localFctB = localFunction(gridFctB_); localFctB.bind(context.element());
      auto localFctC = localFunction(gridFctC_); localFctC.bind(context.element());

      auto const& localFE = rowNode.finiteElement();
      std::size_t numLocalFe = localFE.size();

      int quadDeg = std::max({this->getDegree(2,coeffOrder(localFctA),rowNode,colNode),
                              this->getDegree(1,coeffOrder(localFctB),rowNode,colNode),
                              this->getDegree(0,coeffOrder(localFctC),rowNode,colNode)});

      using QuadratureRules = Dune::QuadratureRules<typename Context::Geometry::ctype, Context::LocalContext::mydimension>;
      auto const& quad = QuadratureRules::rule(context.type(), quadDeg);

      NodeQuadCache<RowNode> cache(localFE.localBasis());
      auto const& shapeGradientsCache = cache.evaluateJacobianAtQP(context, quad);
      auto const& shapeValuesCache = cache.evaluateFunctionAtQP(context, quad);

      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(quad[iq].position()) * quad[iq].weight();

        // the values of the shape functions on the reference element at the quadrature point
        auto const& shapeValues = shapeValuesCache[iq];

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using WorldVector = FieldVector<RangeFieldType,Context::dow>;
        thread_local std::vector<WorldVector> gradients;
        gradients.resize(shapeGradients.size());

        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        const auto A = localFctA(local);
        WorldVector b = makeB(localFctB(local));
        const auto c = localFctC(local);

        IF_CONSTEXPR(conserving) {
          WorldVector gradAi;
          for (std::size_t i = 0; i < numLocalFe; ++i) {
            const auto local_i = rowNode.localIndex(i);
            gradAi = A * gradients[i];
            auto gradBi = b * gradients[i];
            for (std::size_t j = 0; j < numLocalFe; ++j) {
              const auto local_j = colNode.localIndex(j);
              elementMatrix[local_i][local_j] += (dot(gradAi, gradients[j]) + (c * shapeValues[i] - gradBi) * shapeValues[j]) * factor;
            }
          }
        } else {
          WorldVector grad_i;
          for (std::size_t i = 0; i < numLocalFe; ++i) {
            const auto local_i = rowNode.localIndex(i);
            grad_i = A * gradients[i];
            grad_i+= b * shapeValues[i];
            for (std::size_t j = 0; j < numLocalFe; ++j) {
              const auto local_j = colNode.localIndex(j);
              elementMatrix[local_i][local_j] += (dot(grad_i, gradients[j]) + c * shapeValues[i] * shapeValues[j]) * factor;
            }
          }
        }
      }

      localFctA.unbind();
      localFctB.unbind();
      localFctC.unbind();
    }


    template <class Context, class Node, class ElementVector>
    void getElementVector(Context const& context,
                          Node const& node,
                          ElementVector& elementVector)
    {
      static_assert(Node::isLeaf,
        "Operator can be applied to Leaf-Nodes only." );

      auto localFctF = localFunction(gridFctF_); localFctF.bind(context.element());

      auto const& localFE = node.finiteElement();
      std::size_t numLocalFe = localFE.size();

      int quad_order = this->getDegree(0,coeffOrder(localFctF),node);

      using QuadratureRules = Dune::QuadratureRules<typename Context::Geometry::ctype, Context::LocalContext::dimension>;
      auto const& quad = QuadratureRules::rule(context.type(), quad_order);

      NodeQuadCache<Node> cache(localFE.localBasis());
      auto const& shapeValuesCache = cache.evaluateFunctionAtQP(context, quad);

      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // the values of the shape functions on the reference element at the quadrature point
        auto const& shapeValues = shapeValuesCache[iq];

        // The multiplicative factor in the integral transformation formula
        const auto factor = localFctF(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        for (std::size_t i = 0; i < numLocalFe; ++i) {
          const auto local_i = node.localIndex(i);
          elementVector[local_i] += shapeValues[i] * factor;
        }
      }

      localFctF.unbind();
    }

  private:

    template <class LF>
    using HasLocalFunctionOrder = decltype( order(std::declval<LF>()) );

    template <class LocalFct>
    int coeffOrder(LocalFct const& localFct)
    {
      using Concept = Dune::Std::is_detected<HasLocalFunctionOrder, LocalFct>;
      return Dune::Hybrid::ifElse(Concept{},
        [&](auto id) { return order(id(localFct)); },
        [] (auto)    { return 0; });
    }

    template <class T, int N>
    static FieldVector<T,dow> makeB(FieldVector<T,N> const& b) { return b; }

    template <class T, int N>
    static FieldVector<T,dow> makeB(FieldVector<T,N>&& b) { return std::move(b); }

    template <class T>
    static FieldVector<T,dow> makeB(FieldVector<T,1> const& b) { return {T(b)}; }

    template <class T>
    static FieldVector<T,dow> makeB(FieldVector<T,1>&& b) { return {T(b)}; }

    template <class T, std::enable_if_t<std::is_arithmetic<T>::value, int> = 0>
    static FieldVector<T,dow> makeB(T b) { return {b}; }

  private:
    GridFctA gridFctA_;
    GridFctB gridFctB_;
    GridFctC gridFctC_;
    GridFctF gridFctF_;
  };

  template <class PreGridFctA, class PreGridFctB, class PreGridFctC, class PreGridFctF, class c>
  struct PreConvectionDiffusionOperator
  {
    static constexpr bool conserving = c::value;
    PreGridFctA gridFctA;
    PreGridFctB gridFctB;
    PreGridFctC gridFctC;
    PreGridFctF gridFctF;
  };

  template <class PreGridFctA, class PreGridFctB, class PreGridFctC, class PreGridFctF, bool conserving = true>
  auto convectionDiffusion(PreGridFctA const& gridFctA, PreGridFctB const& gridFctB,
                           PreGridFctC const& gridFctC, PreGridFctF const& gridFctF,
                           bool_t<conserving> = {})
  {
    using Pre = PreConvectionDiffusionOperator<PreGridFctA, PreGridFctB, PreGridFctC, PreGridFctF, bool_t<conserving>>;
    return Pre{gridFctA, gridFctB, gridFctC, gridFctF};
  }

  template <class Context, class... T, class GridView>
  auto makeLocalOperator(PreConvectionDiffusionOperator<T...> pre, GridView const& gridView)
  {
    using Pre = PreConvectionDiffusionOperator<T...>;
    auto gridFctA = makeGridFunction(std::move(pre.gridFctA), gridView);
    auto gridFctB = makeGridFunction(std::move(pre.gridFctB), gridView);
    auto gridFctC = makeGridFunction(std::move(pre.gridFctC), gridView);
    auto gridFctF = makeGridFunction(std::move(pre.gridFctF), gridView);

    using GridFctOp = ConvectionDiffusionOperator<Context,
      decltype(gridFctA), decltype(gridFctB), decltype(gridFctC), decltype(gridFctF), Pre::conserving>;

    GridFctOp localOperator{std::move(gridFctA), std::move(gridFctB), std::move(gridFctC), std::move(gridFctF)};
    return localOperator;
  }

  /** @} **/

} // end namespace AMDiS
