#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/StaticSize.hpp>
#include <amdis/utility/LocalBasisCache.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct gradtest {};
  }


  /// first-order operator \f$ \langle\nabla\psi, b\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::gradtest, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::gradtest, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    static const int dow = ContextGeometry<LocalContext>::dow;
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == dow, "Expression must be of vector type." );

  public:
    GridFunctionOperator(tag::gradtest, GridFct const& expr)
      : Super(expr, 1)
    {}

    template <class Context, class Node, class ElementVector>
    void getElementVector(Context const& context,
                          Node const& node,
                          ElementVector& elementVector)
    {
      static_assert(Node::isLeaf, "Node must be Leaf-Node.");

      auto const& quad = this->getQuadratureRule(context.type(), node);
      auto const& localFE = node.finiteElement();
      std::size_t feSize = localFE.size();

      using RangeFieldType = typename NodeQuadCache<Node>::Traits::RangeFieldType;
      NodeQuadCache<Node> cache(localFE.localBasis());

      auto const& shapeGradientsCache = cache.evaluateJacobianAtQP(context, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local);
        const auto dx = context.integrationElement(quad[iq].position()) * quad[iq].weight();

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using WorldVector = FieldVector<RangeFieldType,Context::dow>;
        thread_local std::vector<WorldVector> gradients;
        gradients.resize(shapeGradients.size());

        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        for (std::size_t i = 0; i < feSize; ++i) {
          const auto local_i = node.localIndex(i);
          elementVector[local_i] += dx * (factor * gradients[i]);
        }
      }

    }
  };

  /** @} **/

} // end namespace AMDiS
