#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct divtestvec_divtrialvec {};
  }


  /// second-order operator \f$ \langle\nabla\cdot\Psi, c\,\nabla\cdot\Phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::divtestvec_divtrialvec, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::divtestvec_divtrialvec, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == 1, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::divtestvec_divtrialvec, GridFct const& expr)
      : Super(expr, 2)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isPower && ColNode::isPower,
        "Operator can be applied to Power-Nodes only.");

      const bool sameFE = std::is_same<FiniteElementType_t<RowNode>, FiniteElementType_t<ColNode>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      if (sameFE && sameNode)
        getElementMatrixOptimized(context, quad, rowNode, colNode, elementMatrix);
      else
        getElementMatrixStandard(context, quad, rowNode, colNode, elementMatrix);
    }

  protected:
    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix)
    {
      static const std::size_t CHILDREN = RowNode::CHILDREN;
      static_assert( RowNode::CHILDREN == ColNode::CHILDREN, "" );

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      using RowFieldType = typename NodeQuadCache<typename RowNode::ChildType>::Traits::RangeFieldType;
      using ColFieldType = typename NodeQuadCache<typename ColNode::ChildType>::Traits::RangeFieldType;

      NodeQuadCache<typename RowNode::ChildType> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<typename ColNode::ChildType> colCache(colLocalFE.localBasis());

      auto const& rowGradientsCache = rowCache.evaluateJacobianAtQP(context, quad);
      auto const& colGradientsCache = colCache.evaluateJacobianAtQP(context, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        // The gradients of the shape functions on the reference element
        auto const& rowShapeGradients = rowGradientsCache[iq];
        auto const& colShapeGradients = colGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using RowWorldVector = FieldVector<RowFieldType,Context::dow>;
        thread_local std::vector<RowWorldVector> rowGradients;
        rowGradients.resize(rowShapeGradients.size());

        for (std::size_t i = 0; i < rowGradients.size(); ++i)
          jacobian.mv(rowShapeGradients[i][0], rowGradients[i]);

        using ColWorldVector = FieldVector<ColFieldType,Context::dow>;
        thread_local std::vector<ColWorldVector> colGradients;
        colGradients.resize(colShapeGradients.size());

        for (std::size_t i = 0; i < colGradients.size(); ++i)
          jacobian.mv(colShapeGradients[i][0], colGradients[i]);

        for (std::size_t i = 0; i < rowSize; ++i) {
          for (std::size_t j = 0; j < colSize; ++j) {
            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = rowNode.child(k).localIndex(i);
              const auto local_kj = colNode.child(k).localIndex(j);
              elementMatrix[local_ki][local_kj] += factor * rowGradients[i][k] * colGradients[j][k];
            }
          }
        }
      }
    }


    template <class Context, class QuadratureRule, class Node, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   Node const& node,
                                   Node const& /*colNode*/,
                                   ElementMatrix& elementMatrix)
    {
      static const std::size_t CHILDREN = Node::CHILDREN;

      auto const& localFE = node.child(0).finiteElement();
      std::size_t size = localFE.size();

      using LeafNode = typename Node::ChildType;
      using RangeFieldType = typename NodeQuadCache<LeafNode>::Traits::RangeFieldType;

      NodeQuadCache<LeafNode> cache(localFE.localBasis());

      auto const& shapeGradientsCache = cache.evaluateJacobianAtQP(context, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        // The gradients of the shape functions on the reference element
        auto const& shapeGradients = shapeGradientsCache[iq];

        // Compute the shape function gradients on the real element
        using WorldVector = FieldVector<RangeFieldType,Context::dow>;
        thread_local std::vector<WorldVector> gradients;
        gradients.resize(shapeGradients.size());

        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        for (std::size_t k = 0; k < CHILDREN; ++k) {
          for (std::size_t i = 0; i < size; ++i) {
            const auto local_ki = node.child(k).localIndex(i);
            const auto value = factor * gradients[i][k];
            elementMatrix[local_ki][local_ki] += value * gradients[i][k];

            for (std::size_t j = i+1; j < size; ++j) {
              const auto local_kj = node.child(k).localIndex(j);
              elementMatrix[local_ki][local_kj] += value * gradients[j][k];
              elementMatrix[local_kj][local_ki] += value * gradients[j][k];
            }
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
