#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_trialvec {};
  }


  /// zero-order operator \f$ \langle\psi, \mathbf{b}\cdot\Phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::test_trialvec, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_trialvec, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    static const int dow = ContextGeometry<LocalContext>::dow;
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == dow, "Expression must be of vector type." );

  public:
    GridFunctionOperator(tag::test_trialvec, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isPower,
        "RowNode must be Leaf-Node and ColNode must be a Power-Node.");

      static const std::size_t CHILDREN = ColNode::CHILDREN;
      static_assert( std::is_same<typename GridFct::Range, Dune::FieldVector<double, int(CHILDREN)>>::value, "" );

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      NodeQuadCache<RowNode> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<typename ColNode::ChildType> colCache(colLocalFE.localBasis());

      auto const& rowShapeValuesCache = rowCache.evaluateFunctionAtQP(context,quad);
      auto const& colShapeValuesCache = colCache.evaluateFunctionAtQP(context,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto b = Super::coefficient(local);

        auto const& rowShapeValues = rowShapeValuesCache[iq];
        auto const& colShapeValues = colShapeValuesCache[iq];

        for (std::size_t i = 0; i < rowSize; ++i) {
          const auto local_i = rowNode.localIndex(i);

          for (std::size_t j = 0; j < colSize; ++j) {
            const auto value = b * (factor * rowShapeValues[i] * colShapeValues[j]);

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_kj = colNode.child(k).localIndex(j);
              elementMatrix[local_i][local_kj] += at(value,k);
            }
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
