#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/common/StaticSize.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct testvec {};
  }


  /// zero-order vector-operator \f$ (\mathbf{b}\cdot\Psi) \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::testvec, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::testvec, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    static const int dow = ContextGeometry<LocalContext>::dow;
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Size_v<typename GridFct::Range> == dow, "Expression must be of vector type." );

  public:
    GridFunctionOperator(tag::testvec, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class Node, class ElementVector>
    void getElementVector(Context const& context,
                          Node const& node,
                          ElementVector& elementVector)
    {
      static_assert(Node::isPower,
        "Operator can be applied to Power-Nodes only.");

      static const std::size_t CHILDREN = Node::CHILDREN;

      auto const& quad = this->getQuadratureRule(context.type(), node);
      auto const& localFE = node.child(0).finiteElement();
      std::size_t size = localFE.size();

      NodeQuadCache<typename Node::ChildType> cache(localFE.localBasis());

      auto const& shapeValuesCache = cache.evaluateFunctionAtQP(context,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto exprValue = Super::coefficient(local);

        auto const& shapeValues = shapeValuesCache[iq];

        for (std::size_t i = 0; i < size; ++i) {
          const auto value = exprValue * (factor * shapeValues[i]);
          for (std::size_t k = 0; k < CHILDREN; ++k) {
            const auto local_ki = node.child(k).localIndex(i);
            elementVector[local_ki] += at(value,k);
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
