#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/utility/LocalBasisCache.hpp>
#include <amdis/common/StaticSize.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct testvec_trialvec {};
  }


  /// zero-order operator \f$ \langle\Psi, c\,\Phi\rangle \f$, or \f$ \langle\Psi, A\,\Phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::testvec_trialvec, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::testvec_trialvec, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    static const int dow = ContextGeometry<LocalContext>::dow;
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    using expr_value_type = typename GridFct::Range;
    static_assert( Size_v<expr_value_type> == 1 || (Rows_v<expr_value_type> == dow && Cols_v<expr_value_type> == dow),
      "Expression must be of scalar or matrix type." );

  public:
    GridFunctionOperator(tag::testvec_trialvec, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isPower && ColNode::isPower,
        "Operator can be applied to Power-Nodes only.");

      static_assert( (RowNode::CHILDREN == ColNode::CHILDREN), "" );
      // theoretically the restriction of the same nr of childs would not be necessary

      const bool sameFE = std::is_same<FiniteElementType_t<RowNode>, FiniteElementType_t<ColNode>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();

      using Category = ValueCategory_t<expr_value_type>;

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      if (sameFE && sameNode && std::is_same<Category,tag::scalar>::value)
        getElementMatrixOptimized(context, quad, rowNode, colNode, elementMatrix, Category{});
      else
        getElementMatrixStandard(context, quad, rowNode, colNode, elementMatrix, Category{});
    }

  protected: // specialization for different value-categories

    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix,
                                  tag::scalar)
    {
      static const std::size_t CHILDREN = RowNode::CHILDREN;

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();

      NodeQuadCache<typename RowNode::ChildType> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<typename ColNode::ChildType> colCache(colLocalFE.localBasis());

      auto const& rowShapeValuesCache = rowCache.evaluateFunctionAtQP(context,quad);
      auto const& colShapeValuesCache = colCache.evaluateFunctionAtQP(context,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        auto const& rowShapeValues = rowShapeValuesCache[iq];
        auto const& colShapeValues = colShapeValuesCache[iq];

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto value = factor * rowShapeValues[i];

          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto value0 = value * colShapeValues[j];

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = rowNode.child(k).localIndex(i);
              const auto local_kj = colNode.child(k).localIndex(j);

              elementMatrix[local_ki][local_kj] += value0;
            }
          }
        }
      }
    }


    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   RowNode const& node,
                                   ColNode const& /*colNode*/,
                                   ElementMatrix& elementMatrix,
                                   tag::scalar)
    {
      static const std::size_t CHILDREN = RowNode::CHILDREN;

      auto const& localFE = node.child(0).finiteElement();

      NodeQuadCache<typename RowNode::ChildType> cache(localFE.localBasis());

      auto const& shapeValuesCache = cache.evaluateFunctionAtQP(context,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(quad[iq].position()) * quad[iq].weight();

        auto const& shapeValues = shapeValuesCache[iq];

        for (std::size_t i = 0; i < localFE.size(); ++i) {
          const auto value = factor * shapeValues[i];

          for (std::size_t k = 0; k < CHILDREN; ++k) {
            const auto local_ki = node.child(k).localIndex(i);
            elementMatrix[local_ki][local_ki] += value * shapeValues[i];
          }

          for (std::size_t j = i+1; j < localFE.size(); ++j) {
            const auto value0 = value * shapeValues[j];

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = node.child(k).localIndex(i);
              const auto local_kj = node.child(k).localIndex(j);

              elementMatrix[local_ki][local_kj] += value0;
              elementMatrix[local_kj][local_ki] += value0;
            }
          }
        }
      }
    }

    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix,
                                  tag::matrix)
    {
      static const std::size_t CHILDREN = RowNode::CHILDREN;
      static_assert( std::is_same<expr_value_type, Dune::FieldMatrix<double, int(CHILDREN), int(CHILDREN)>>::value, "" );

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();

      NodeQuadCache<typename RowNode::ChildType> rowCache(rowLocalFE.localBasis());
      NodeQuadCache<typename ColNode::ChildType> colCache(colLocalFE.localBasis());

      auto const& rowShapeValuesCache = rowCache.evaluateFunctionAtQP(context,quad);
      auto const& colShapeValuesCache = colCache.evaluateFunctionAtQP(context,quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(quad[iq].position()) * quad[iq].weight();
        const Dune::FieldMatrix<double, int(CHILDREN), int(CHILDREN)> exprValue = Super::coefficient(local);

        auto const& rowShapeValues = rowShapeValuesCache[iq];
        auto const& colShapeValues = colShapeValuesCache[iq];

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto value0 = factor * rowShapeValues[i];

          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto value = value0 * colShapeValues[j];
            const auto mat = exprValue * value;

            for (std::size_t k0 = 0; k0 < CHILDREN; ++k0) {
              const auto local_ki = rowNode.child(k0).localIndex(i);
              for (std::size_t k1 = 0; k1 < CHILDREN; ++k1) {
                const auto local_kj = colNode.child(k1).localIndex(j);

                elementMatrix[local_ki][local_kj] += mat[k0][k1];
              }
            }
          }
        }
      }
    }

    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   RowNode const& node,
                                   ColNode const& /*colNode*/,
                                   ElementMatrix& elementMatrix,
                                   tag::matrix)
    {
      assert(false && "Not implemented.");
    }
  };

  /** @} **/

} // end namespace AMDiS
