#pragma once

#include <array>
#include <functional>
#include <type_traits>
#include <utility>

#include <dune/common/indices.hh>
#include <dune/common/tuplevector.hh>

#include <dune/typetree/treepath.hh>

#include <amdis/common/Apply.hpp>

// NOTE: backport of dune/typetree/treecontainer.hh

namespace AMDiS
{
  namespace Impl
  {
    /*
      * \brief A factory class creating a hybrid container compatible with a type tree
      *
      * This class allows to create a nested hybrid container having the same structure
      * as a given type tree. Power nodes are represented as std::array's while composite
      * nodes are represented as Dune::TupleVector's. The stored values for the leaf nodes
      * are creating using a given predicate. Once created, the factory provides an
      * operator() creating the container for the tree given as argument.
      *
      * \tparam LeafToValue Type of a predicate that determines the stored values at the leafs
      */
    template<class LeafToValue>
    class ContainerFactory
    {
    public:

      /**
        * \brief Create ContainerFactory
        *
        * The given predicate will be stored by value.
        *
        * \param A predicate used to generate the stored values for the leaves
        */
      ContainerFactory(LeafToValue leafToValue) :
        leafToValue_(leafToValue)
      {}

      template<class Node,
        std::enable_if_t<Node::isLeaf, int> = 0>
      auto operator()(const Node& node)
      {
        return leafToValue_(node);
      }

      template<class Node,
        std::enable_if_t<Node::isPower, int> = 0>
      auto operator()(const Node& node)
      {
        using TransformedChild = decltype((*this)(node.child(0)));
        return std::array<TransformedChild, Node::degree()>();
      }

      template<class Node,
        std::enable_if_t<Node::isComposite, int> = 0>
      auto operator()(const Node& node)
      {
        return Tools::apply_indices([&](auto... indices) { return Dune::makeTupleVector((*this)(node.child(indices))...); },
                                    index_t<Node::degree()>{});
      }

    private:
      LeafToValue leafToValue_;
    };


    /*
      * \brief Wrap nested container to provide a VectorBackend
      */
    template<class Container>
    class TreeContainerVectorBackend
    {
      template<class C>
      static constexpr decltype(auto) accessByTreePath(C&& container, const Dune::TypeTree::HybridTreePath<>& path)
      {
        return container;
      }

      template<class C, class... T>
      static constexpr decltype(auto) accessByTreePath(C&& container, const Dune::TypeTree::HybridTreePath<T...>& path)
      {
        auto head = Dune::TypeTree::treePathEntry(path,Dune::Indices::_0);
        auto tailPath = Tools::apply_indices([&](auto... i)
        {
          using namespace Dune::TypeTree;
          return hybridTreePath(treePathEntry(path,index_t<i.value+1>{})...);
        }, index_t<sizeof...(T)-1>{});
        return accessByTreePath(container[head], tailPath);
      }

    public:
      TreeContainerVectorBackend(Container&& container) :
        container_(std::move(container))
      {}

      TreeContainerVectorBackend(TreeContainerVectorBackend&& other) :
        container_(std::move(other.container_))
      {}

      template<class... T>
      decltype(auto) operator[](const Dune::TypeTree::HybridTreePath<T...>&  path) const
      {
        return accessByTreePath(container_, path);
      }

      template<class... T>
      decltype(auto) operator[](const Dune::TypeTree::HybridTreePath<T...>&  path)
      {
        return accessByTreePath(container_, path);
      }

      const Container& data() const
      {
        return container_;
      }

      Container& data()
      {
        return container_;
      }

    private:
      Container container_;
    };

    template<class Container>
    auto makeTreeContainerVectorBackend(Container&& container)
    {
      return TreeContainerVectorBackend<std::decay_t<Container>>(std::forward<Container>(container));
    }

  } // end namespace Impl

    /** \addtogroup TypeTree
     *  \{
     */

    /**
     * \brief Create container havin the same structure as the given tree
     *
     * This class allows to create a nested hybrid container having the same structure
     * as a given type tree. Power nodes are represented as std::array's while composite
     * nodes are represented as Dune::TupleVector's. The stored values for the leaf nodes
     * are creating using a given predicate. For convenience the created container is
     * not returned directly. Instead, the returned object stores the container and
     * provides operator[] access using a HybridTreePath.
     *
     * \param tree The tree which should be mapper to a container
     * \param leafToValue A predicate used to generate the stored values for the leaves
     *
     * \returns A container matching the tree structure
     */
    template<class Tree, class LeafToValue>
    auto makeTreeContainer(const Tree& tree, LeafToValue&& leafToValue)
    {
      auto f = std::ref(leafToValue);
      auto factory = Impl::ContainerFactory<decltype(f)>(f);
      return Impl::makeTreeContainerVectorBackend(factory(tree));
    }

    /**
     * \brief Create container havin the same structure as the given tree
     *
     * This class allows to create a nested hybrid container having the same structure
     * as a given type tree. Power nodes are represented as std::array's while composite
     * nodes are represented as Dune::TupleVector's. The stored values for the leaf nodes
     * are of the given type Value. For convenience the created container is
     * not returned directly. Instead, the returned object stores the container and
     * provides operator[] access using a HybridTreePath.
     *
     * \tparam Value Type of the values to be stored for the leafs. Should be default constructible.
     * \param leafToValue A predicate used to generate the stored values for the leaves
     *
     * \returns A container matching the tree structure
     */
    template<class Value, class Tree>
    auto makeTreeContainer(const Tree& tree)
    {
      return makeTreeContainer(tree, [](const auto&) {return Value{};});
    }

    /**
     * \brief Alias to container type generated by makeTreeContainer for given value and tree type
     */
    template<class Value, class Tree>
    using TreeContainer = std::decay_t<decltype(makeTreeContainer<Value>(std::declval<const Tree&>()))>;

    //! \} group TypeTree

} //namespace AMDiS
