#pragma once

#include <dune/typetree/visitor.hh>
#include <amdis/typetree/Traversal.hpp>

namespace AMDiS
{
  // from dune-typetree merge-request !2
  namespace Impl
  {
    template <class PreFunc, class LeafFunc, class PostFunc>
    class CallbackVisitor
        : public Dune::TypeTree::TreeVisitor
    {
    public:
      CallbackVisitor(PreFunc& preFunc, LeafFunc& leafFunc, PostFunc& postFunc)
        : preFunc_(preFunc)
        , leafFunc_(leafFunc)
        , postFunc_(postFunc)
      {}

      template <typename Node, typename TreePath>
      void pre(Node&& node, TreePath treePath)
      {
        preFunc_(node, treePath);
      }

      template <typename Node, typename TreePath>
      void leaf(Node&& node, TreePath treePath)
      {
        leafFunc_(node, treePath);
      }

      template <typename Node, typename TreePath>
      void post(Node&& node, TreePath treePath)
      {
        postFunc_(node, treePath);
      }

    private:
      PreFunc& preFunc_;
      LeafFunc& leafFunc_;
      PostFunc& postFunc_;
    };

    template <class PreFunc, class LeafFunc, class PostFunc>
    auto callbackVisitor(PreFunc& preFunc, LeafFunc& leafFunc, PostFunc& postFunc)
    {
      return CallbackVisitor<PreFunc, LeafFunc, PostFunc>(preFunc, leafFunc, postFunc);
    }

  } // namespace Impl


  /**
    * \brief Traverse tree and visit each node
    *
    * All passed callback functions are called with the
    * node and corresponding treepath as arguments.
    *
    * \param tree The tree to traverse
    * \param preFunc This function is called for each inner node before visiting its children
    * \param leafFunc This function is called for each leaf node
    * \param postFunc This function is called for each inner node after visiting its children
    */
  template <class Tree, class PreFunc, class LeafFunc, class PostFunc>
  void forEachNode_(Tree&& tree, PreFunc&& preFunc, LeafFunc&& leafFunc, PostFunc&& postFunc)
  {
    traverseTree(tree, Impl::callbackVisitor(preFunc, leafFunc, postFunc));
  }

  /**
    * \brief Traverse tree and visit each node
    *
    * All passed callback functions are called with the
    * node and corresponding treepath as arguments.
    *
    * \param tree The tree to traverse
    * \param innerFunc This function is called for each inner node before visiting its children
    * \param leafFunc This function is called for each leaf node
    */
  template <class Tree, class InnerFunc, class LeafFunc>
  void forEachNode_(Tree&& tree, InnerFunc&& innerFunc, LeafFunc&& leafFunc)
  {
    auto nop = [](auto&&... args) {};
    forEachNode_(tree, innerFunc, leafFunc, nop);
  }

  /**
    * \brief Traverse tree and visit each node
    *
    * The passed callback function is called with the
    * node and corresponding treepath as arguments.
    *
    * \param tree The tree to traverse
    * \param nodeFunc This function is called for each node
    */
  template <class Tree, class NodeFunc>
  void forEachNode_(Tree&& tree, NodeFunc&& nodeFunc)
  {
    forEachNode_(tree, nodeFunc, nodeFunc);
  }

  /**
    * \brief Traverse tree and visit each leaf node
    *
    * The passed callback function is called with the
    * node and corresponding treepath as arguments.
    *
    * \param tree The tree to traverse
    * \param leafFunc This function is called for each leaf node
    */
  template <class Tree, class LeafFunc>
  void forEachLeafNode_(Tree&& tree, LeafFunc&& leafFunc)
  {
    auto nop = [](auto&&... args) {};
    forEachNode_(tree, nop, leafFunc, nop);
  }

} // end namespace AMDiS
